# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Class for Rayleigh opacties. Only for H2 and He at the moment. 
"""
import numpy as np
from .util.cst import PI, KBOLTZ
from .util.singleton import Singleton

class Rayleigh(Singleton):
    """Class to access Rayleigh opacities
    """

    def init(self, *args, **kwds):
        """Initializes various parameters for Rayleigh computations
        """
        N_std=1.e5/(KBOLTZ*273.15)
        self._mult_factor=24.*PI**3/(N_std)**2
        self._mult_factor=self._mult_factor*100.**4
        # last 100.**4 is because we are using wavenumbers in cm^-1
        # instead of wavelength in m (see eq 12 of Caldas 2019)
        self._ABC={'He':np.array([0.2283])}

    def sigma(self, wns, vmr):
        """Computes the Rayleigh cross section for the gas.
        This one is faster than sigma_array, but can be used only when vmr values are floats.

        Parameters
        ----------
            wns: array
                array of wavenumbers

            vmr: dict of floats
                Keys are molecule names. Values are the volume mixing ratios

        Returns
        -------
            array of shape (wns.size)
                Rayleigh cross section for the whole gas in m^2/molecule
        """
        res=np.zeros(wns.size)
        wave = 1e8/wns
        for mol, x in vmr.items():
            to_add=True
            if mol == 'H2':
                tmp=((8.14E-13)*(wave**(-4.))* \
                    (1+(1.572E6)*(wave**(-2.))+(1.981E12)*(wave**(-4.))))*1E-4
                #res+=x*(8.14e-33+1.28e-42*wns**2+1.61e-51*wns**4)*wns**4/1.e16
            elif mol == 'He':
                tmp=((5.484E-14)*(wave**(-4.))*(1+(2.44E5)*(wave**(-2.))))*1E-4
            else:
                to_add=False
            if to_add: res+=x*tmp

        return res

    def sigma_array(self, wns, vmr):
        """Computes the Rayleigh cross section for the gas.

        Parameters
        ----------
            wns: array
                array of wavenumbers

            vmr: dict of arrays
                Keys are molecule names. Values are arrays the volume mixing ratios

        Returns
        -------
            array of shape (vmr.values.size, wns.size)
                Rayleigh cross section for the whole gas in m^2/molecule
        """
        first_mol=True
        wave = 1e8/wns
        for mol, x in vmr.items():
            to_add=True
            x_array=np.array(x)
            if first_mol:
                res=np.zeros((x_array.size,wns.size))
                first_mol=False
            if mol == 'H2':
                tmp=((8.14E-13)*(wave**(-4.))* \
                    (1+(1.572E6)*(wave**(-2.))+(1.981E12)*(wave**(-4.))))*1E-4
                #res+=x*(8.14e-33+1.28e-42*wns**2+1.61e-51*wns**4)*wns**4/1.e16
            elif mol == 'He':
                tmp=((5.484E-14)*(wave**(-4.))*(1+(2.44E5)*(wave**(-2.))))*1E-4
            else:
                to_add=False
            if to_add: res+=x_array[:,None]*tmp

        return res



# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

Module containing a class to handle a database of CIA tables to compute opacities with it.
"""
import numpy as np

from .cia_table import Cia_table

class CIAdatabase(object):
    """Class to group :class:`~exo_k.cia_table.Cia_table` objects
    and combine them in radiative transfer
    """

    def __init__(self, filenames, *str_filters, remove_zeros=True, **kwargs):
        """Initializes cia tables and supporting data from a list of filenames.

        Parameters
        ----------
            filenames: list
                List of names (not full path) of the input cia files. 
                The files must be in the global search path.
                
        A local search path can be specified with 'search_path='

        See the options of :class:`~exo_k.cia_table.Cia_table` __init__ method.
        """
        self.cia_tables={}
        self.wns=None
        self.Nw=None
        self.abs_coeff_unit=None
        if filenames is not None:
            for filename in filenames:
                tmp_cia_table=Cia_table(*([filename]+list(str_filters)),
                    remove_zeros=remove_zeros, **kwargs)
                self.add_cia_tables(tmp_cia_table)
    
    def add_cia_tables(self, *cia_tables):
        """Adds new :class:`~exo_k.cia_table.Cia_table` objects to a CIA database.

        Parameters
        ----------
            cia_tables: :class:`~exo_k.cia_table.Cia_table`
                As many cia tables as you want.
        """
        for cia_table in cia_tables:
            if self.abs_coeff_unit is None:
                self.abs_coeff_unit=cia_table.abs_coeff_unit
            elif self.abs_coeff_unit!=cia_table.abs_coeff_unit:
                print('CIAdatabase units for cia coeff: {k}'.format(k=self.abs_coeff_unit))
                print('{mol1}-{mol2} units for cia coeff: {k}'.format(\
                    mol1=cia_table.mol1, mol2=cia_table.mol2, k=cia_table.abs_coeff_unit))
                raise RuntimeError("""You naughty:
                all CIAdatabase in a database must have the same coeff units""")

            if cia_table.mol1 in self.cia_tables:
                if cia_table.mol2 in self.cia_tables[cia_table.mol1]:
                    continue
                else:
                    self.cia_tables[cia_table.mol1][cia_table.mol2]=cia_table
            else:
                self.cia_tables[cia_table.mol1]={cia_table.mol2:cia_table}

    def __getitem__(self, molecule):
        """Overrides getitem so that CIAdatabase['mol'] directly accesses 
        the database for that molecule.
        """
        if molecule not in self.cia_tables:
            raise KeyError('The requested molecule is not available.')
        return self.cia_tables[molecule]

    def copy(self):
        """Creates a new instance of :class:`CIAdatabase` object and (deep) copies data into it
        """
        res=CIAdatabase(None)
        for dic_cia_tabs in self.cia_tables.values():
            for cia_tab in dic_cia_tabs.values():
                res.add_cia_tables(cia_tab.copy())
        return res

    def __repr__(self):
        """Method to output
        """
        output='The available molecule pairs are: \n'
        for mol1, dico in self.cia_tables.items():
            for mol2, cia_tab in dico.items():
                output+=mol1+'-'+mol2+'->'+cia_tab.filename+'\n'
        if self.wns is not None:
            output+='All tables share a common spectral grid\n'
        else:
            output+='All tables do NOT have common spectral grid\n'
            output+='You will need to run sample before using the database\n'

        return output

    def sample(self, wngrid, remove_zeros=False):
        """Samples all the cia_table in the database on the same wavenumber grid
        to be able to use them in radiative transfer modules.

        Parameters
        ----------
            wngrid : array
                new wavenumber grid (cm-1)
        """
        for mol1 in self.cia_tables.values():
            for cia_table in mol1.values():
                cia_table.sample(wngrid,remove_zeros)
        self.wns=np.array(wngrid)
        self.Nw=self.wns.size

    def convert_to_mks(self):
        """Converts units of all Cia_tables to MKS.
        """
        first=True
        for mol1 in self.cia_tables.values():
            for cia_table in mol1.values():
                cia_table.convert_to_mks()
                if first:
                    self.abs_coeff_unit=cia_table.abs_coeff_unit
        

    def cia_cross_section(self, logP_array, T_array, gas_comp, wngrid_limit=None):
        """Computes the absorption coefficient in m^-1 for the whole mix specified 
        (assumes data in MKS).

        Parameters
        ----------
            logP_array: array

            T_array: array
                log10 Pressure and temperature profiles

            gas_comp: :class:`~exo_k.gas_mix.Gas_mix` object
                behaves like a dict with mol names as keys and vmr as values.

            wngrid_limit: array, optional
                Smaller and bigger wavenumbers inside which to perform the calculation.

        Returns:
            array:
                The cia effective cross section coefficient profile for the whole gas (in m^2).
        """
        logP_array=np.array(logP_array)
        T_array=np.array(T_array)
        first=True
        if self.wns is None: raise RuntimeError("""cia tables must be sampled
            on the same grid before calling cia_cross_section.
            Please use the sample() method.""")
        for mol,x1 in gas_comp.items():
            if mol in self.cia_tables.keys():
                for mol2 in self.cia_tables[mol].keys():
                    if mol2 in gas_comp.keys():
                        x2=gas_comp[mol2]
                        if first:
                            res=self.cia_tables[mol][mol2].effective_cross_section( \
                                logP_array, T_array, x1, x2, wngrid_limit=wngrid_limit)
                            first=False
                        else:
                            res+=self.cia_tables[mol][mol2].effective_cross_section( \
                                logP_array, T_array, x1, x2, wngrid_limit=wngrid_limit)
        if first: # means that no molecule was in the database, we need to initialize res
            res=np.zeros((T_array.size,self.wns.size))
        return res

        




# -*- coding: utf-8 -*-
"""
@author: jeremy leconte

A class based on a singleton to store global options only once for every instance. 
"""
import os.path
from glob import glob
from fnmatch import fnmatch,fnmatchcase
from .util.singleton import Singleton

class Settings(Singleton):
    """A class based on a singleton to store global options only once for every instance.

    So all the following methods can be called using the following syntax:

    >>> exo_k.Settings().method_name(agrs)

    In gerneal, they will change internal global attributes
    that change the global behavior of some routines in the library. 
    """

    def init(self, *args, **kwds):
        self.reset_search_path()
        self._log_interp = True
        self._convert_to_mks = False
        self._delimiter = '_'
        self._case_sensitive = False

    def reset_search_path(self):
        """Set default search path.
        """
        self._search_path = [os.path.abspath('.')]

    def add_search_path(self, *search_paths):
        """Add path(s) to the list of paths that will be searched for
        correlated-k and x-sec files.

        Parameters
        ----------
            search_path : string or list of strings
                Search path(s) to look for opacities.

        Examples
        --------

            >>> exo_k.Settings().add_search_path('data/xsec','data/corrk','data/cia')
            >>> exo_k.Settings().search_path()
            ['/Users/jleconte/atmosphere/RadiativeTransfer/exo_k',
             '/Users/jleconte/atmosphere/RadiativeTransfer/exo_k/data/xsec',
             '/Users/jleconte/atmosphere/RadiativeTransfer/exo_k/data/corrk',
             '/Users/jleconte/atmosphere/RadiativeTransfer/exo_k/data/cia']
        """
        for path in search_paths:
            if not os.path.isdir(path):
                raise NotADirectoryError("""The search_path you provided
                    does not exist or is not a directory""")
            else:
                if os.path.abspath(path) not in self._search_path:
                    self._search_path.append(os.path.abspath(path))

    def set_search_path(self, *search_paths):
        """Sets the path(s) that will be searched for correlated-k and x-sec files .

        Parameters
        ----------
            search_path : string or list of strings
                Search path(s) to look for opacities.
        """
        if not os.path.isdir(search_paths[0]):
            raise NotADirectoryError("""The search_path you provided
                does not exist or is not a directory""")
        else:
            self._search_path=[os.path.abspath(search_paths[0])]
        if len(search_paths)>1: self.add_search_path(*search_paths[1:])

    def search_path(self):
        """Returns the current value of the global search path (_search_path)
        """
        return self._search_path

    def set_delimiter(self, newdelimiter):
        """Sets the delimiter string used to separate molecule names in filenames.

        Parameters
        ----------
            newdelimiter: string
                New delimiter to use. Default is '_'.

        Exmaple
        -------
            If I have a file named 'H2O.R10000_xsec.hdf5'
            that I want to load in a `Kdatabase`, the default
            settings will result in an error:

            >>> database=xk.Kdatabase(['H2O'],'R10000')
             No file was found with these filters: 
             ('H2O_', 'R1000') in the following directories:
             ['/home/falco/xsec/xsec_sampled_R10000_0.3-15'] 

            Using

            >>> xk.Settings().set_delimiter('.')
            >>> database=xk.Kdatabase(['H2O'],'R10000')

            finds the file.
        """
        self._delimiter = newdelimiter

    def set_log_interp(self, log_interp):
        """Sets the default interpolation mode for kdata. Default is Log. 

        Parameters
        ----------
            log_interp: boolean
                If True, log interpolation. Linear if False.
        """
        self._log_interp = log_interp

    def set_case_sensitive(self, case_sensitive):
        """Set whether name matching is case sensitive. Default is False.

        Parameters
        ----------
            case_sensitive: boolean
                If True, name matching is case sensitive.
        """
        self._case_sensitive = case_sensitive

    def set_mks(self, set_mks):
        """Forces conversion to mks system.
        
        Parameters
        ----------
            set_mks: boolean
                If True, all datasets are converted to mks upon loading.
        """
        self._convert_to_mks = set_mks

    def list_files(self, *str_filters, only_one = False, search_path = None):
        """A routine that provides a list of all filenames containing
        a set of string filters in the global _search_path or a local one.

        Parameters
        ----------
            *str_filters: str
                A set of strings that need to be contained in the name of the file
            only_one: boolean, optional
                If true, only one filename is returned (the first one).
                If false, a list is returned. Default is False.
            search_path: str, optional
                If search_path is provided, it locally overrides
                the global _search_path settings
                and only files in search_path are returned.

        Returns
        -------
            list of strings
                List of filenames corresponding to all the str_filters
        """
        local_search_path=self._search_path
        if search_path is not None:
            local_search_path=[search_path]

        filenames = [f for path in local_search_path for f in glob(os.path.join(path,'*'))]
        finalnames=filenames[:]
        for filename in filenames:
            if os.path.isdir(filename):
                finalnames.remove(filename)
                continue
            if self._case_sensitive:
                for filt in str_filters:
                    if not fnmatchcase(os.path.basename(filename),'*'+filt+'*'):
                        finalnames.remove(filename)
                        break
            else:
                for filt in str_filters:
                    if not fnmatch(os.path.basename(filename.lower()),'*'+filt.lower()+'*'):
                        finalnames.remove(filename)
                        break
        if len(finalnames)>1 and only_one: 
            print("""be careful: {filt}
            String filters not specific enough, several corresponding files have been found.
            We will use the first one:
            {file}
            Other files are:
            {other_files}""".format(filt=str_filters,file=finalnames[0], \
                other_files=finalnames[1:]))
            finalnames=[finalnames[0]]
        if not finalnames:
            print("""No file was found with these filters: 
            {filt}
            in the following directories:
            {path}
            """.format(filt=str_filters,path=local_search_path))
            raise RuntimeError()
        # an empty sequence yields False in a conditional statement ! 
        return finalnames

